use anyhow::bail;
use anyhow::Context;
use anyhow::Result;
use clap::Parser;
use clap::Subcommand;
use daemon::bdk::bitcoin;
use daemon::bdk::bitcoin::Address;
use daemon::bdk::bitcoin::Amount;
use daemon::bdk::FeeRate;
use daemon::connection::connect;
use daemon::db;
use daemon::monitor;
use daemon::oracle;
use daemon::projection;
use daemon::seed::RandomSeed;
use daemon::seed::Seed;
use daemon::seed::UmbrelSeed;
use daemon::wallet;
use daemon::TakerActorSystem;
use daemon::HEARTBEAT_INTERVAL;
use daemon::N_PAYOUTS;
use model::olivia;
use model::Identity;
use model::SETTLEMENT_INTERVAL;
use shared_bin::fairings;
use shared_bin::logger;
use shared_bin::logger::LevelFilter;
use std::net::SocketAddr;
use std::path::PathBuf;
use std::time::Duration;
use tokio_tasks::Tasks;
use xtra::Actor;

mod routes;

pub const ANNOUNCEMENT_LOOKAHEAD: time::Duration = time::Duration::hours(24);

const MAINNET_ELECTRUM: &str = "ssl://blockstream.info:700";
const MAINNET_MAKER: &str = "mainnet.itchysats.network:10000";
const MAINNET_MAKER_ID: &str = "7e35e34801e766a6a29ecb9e22810ea4e3476c2b37bf75882edf94a68b1d9607";

const TESTNET_ELECTRUM: &str = "ssl://blockstream.info:993";
const TESTNET_MAKER: &str = "testnet.itchysats.network:9999";
const TESTNET_MAKER_ID: &str = "69a42aa90da8b065b9532b62bff940a3ba07dbbb11d4482c7db83a7e049a9f1e";

#[derive(Parser)]
struct Opts {
    /// The IP address or hostname of the other party (i.e. the maker).
    ///
    /// If not specified it defaults to the itchysats maker for the mainnet or testnet.
    #[clap(long)]
    maker: Option<String>,

    /// The public key of the maker as a 32 byte hex string.
    ///
    /// If not specified it defaults to the itchysats maker-id for mainnet or testnet.
    #[clap(long, parse(try_from_str = parse_x25519_pubkey))]
    maker_id: Option<x25519_dalek::PublicKey>,

    /// The IP address to listen on for the HTTP API.
    #[clap(long, default_value = "127.0.0.1:8000")]
    http_address: SocketAddr,

    /// Where to permanently store data, defaults to the current working directory.
    #[clap(long)]
    data_dir: Option<PathBuf>,

    /// If enabled logs will be in json format
    #[clap(short, long)]
    json: bool,

    /// Configure the log level, e.g.: one of Error, Warn, Info, Debug, Trace
    #[clap(short, long, default_value = "Debug")]
    log_level: LevelFilter,

    /// Password for the web interface.
    ///
    /// If not provided, will be derived from the seed.
    #[clap(long)]
    password: Option<rocket_basicauth::Password>,

    #[clap(subcommand)]
    network: Option<Network>,

    #[clap(short, long, parse(try_from_str = parse_umbrel_seed))]
    umbrel_seed: Option<[u8; 32]>,
}

impl Opts {
    fn network(&self) -> Network {
        self.network.clone().unwrap_or_else(|| Network::Mainnet {
            electrum: MAINNET_ELECTRUM.to_string(),
            withdraw: None,
        })
    }

    fn maker(&self) -> Result<(String, x25519_dalek::PublicKey)> {
        let network = self.network();

        let maker_url = match self.maker.clone() {
            Some(maker) => maker,
            None => match network {
                Network::Mainnet { .. } => MAINNET_MAKER.to_string(),
                Network::Testnet { .. } => TESTNET_MAKER.to_string(),
                Network::Signet { .. } => bail!("No maker default URL configured for signet"),
            },
        };

        let maker_id = match self.maker_id {
            Some(maker_id) => maker_id,
            None => match network {
                Network::Mainnet { .. } => parse_x25519_pubkey(MAINNET_MAKER_ID)?,
                Network::Testnet { .. } => parse_x25519_pubkey(TESTNET_MAKER_ID)?,
                Network::Signet { .. } => {
                    bail!("No maker default public key configured for signet")
                }
            },
        };

        Ok((maker_url, maker_id))
    }
}

fn parse_x25519_pubkey(s: &str) -> Result<x25519_dalek::PublicKey> {
    let mut bytes = [0u8; 32];
    hex::decode_to_slice(s, &mut bytes)?;
    Ok(x25519_dalek::PublicKey::from(bytes))
}

fn parse_umbrel_seed(s: &str) -> Result<[u8; 32]> {
    let mut bytes = [0u8; 32];
    hex::decode_to_slice(s, &mut bytes)?;
    Ok(bytes)
}

#[derive(Parser, Clone)]
enum Network {
    /// Run on mainnet (default)
    Mainnet {
        /// URL to the electrum backend to use for the wallet.
        #[clap(long, default_value = MAINNET_ELECTRUM)]
        electrum: String,

        #[clap(subcommand)]
        withdraw: Option<Withdraw>,
    },
    /// Run on testnet
    Testnet {
        /// URL to the electrum backend to use for the wallet.
        #[clap(long, default_value = TESTNET_ELECTRUM)]
        electrum: String,

        #[clap(subcommand)]
        withdraw: Option<Withdraw>,
    },
    /// Run on signet
    Signet {
        /// URL to the electrum backend to use for the wallet.
        #[clap(long)]
        electrum: String,

        #[clap(subcommand)]
        withdraw: Option<Withdraw>,
    },
}

#[derive(Subcommand, Clone)]
enum Withdraw {
    Withdraw {
        /// Optionally specify the amount of Bitcoin to be withdrawn. If not specified the wallet
        /// will be drained. Amount is to be specified with denomination, e.g. "0.1 BTC"
        #[clap(long)]
        amount: Option<Amount>,
        /// Optionally specify the fee-rate for the transaction. The fee-rate is specified as sats
        /// per vbyte, e.g. 5.0
        #[clap(long)]
        fee: Option<f32>,
        /// The address to receive the Bitcoin.
        #[clap(long)]
        address: Address,
    },
}

impl Network {
    fn electrum(&self) -> &str {
        match self {
            Network::Mainnet { electrum, .. } => electrum,
            Network::Testnet { electrum, .. } => electrum,
            Network::Signet { electrum, .. } => electrum,
        }
    }

    fn bitcoin_network(&self) -> bitcoin::Network {
        match self {
            Network::Mainnet { .. } => bitcoin::Network::Bitcoin,
            Network::Testnet { .. } => bitcoin::Network::Testnet,
            Network::Signet { .. } => bitcoin::Network::Signet,
        }
    }

    fn data_dir(&self, base: PathBuf) -> PathBuf {
        match self {
            Network::Mainnet { .. } => base.join("mainnet"),
            Network::Testnet { .. } => base.join("testnet"),
            Network::Signet { .. } => base.join("signet"),
        }
    }

    fn withdraw(&self) -> &Option<Withdraw> {
        match self {
            Network::Mainnet { withdraw, .. } => withdraw,
            Network::Testnet { withdraw, .. } => withdraw,
            Network::Signet { withdraw, .. } => withdraw,
        }
    }
}

#[rocket::main]
async fn main() -> Result<()> {
    let opts = Opts::parse();

    let network = opts.network();
    let (maker_url, maker_id) = opts.maker()?;

    logger::init(opts.log_level, opts.json).context("initialize logger")?;
    tracing::info!("Running version: {}", daemon::version::version());
    let settlement_interval_hours = SETTLEMENT_INTERVAL.whole_hours();

    tracing::info!(
        "CFDs created with this release will settle after {settlement_interval_hours} hours"
    );

    let data_dir = opts
        .data_dir
        .clone()
        .unwrap_or_else(|| std::env::current_dir().expect("unable to get cwd"));

    let data_dir = network.data_dir(data_dir);

    if !data_dir.exists() {
        tokio::fs::create_dir_all(&data_dir).await?;
    }

    let maker_identity = Identity::new(maker_id);

    let bitcoin_network = network.bitcoin_network();
    let (ext_priv_key, identities, web_password) = match opts.umbrel_seed {
        Some(seed_bytes) => {
            let seed = UmbrelSeed::from(seed_bytes);
            let ext_priv_key = seed.derive_extended_priv_key(bitcoin_network)?;
            let identities = seed.derive_identities();
            let web_password = opts.password.unwrap_or_else(|| seed.derive_auth_password());
            (ext_priv_key, identities, web_password)
        }
        None => {
            let seed = RandomSeed::initialize(&data_dir.join("taker_seed")).await?;
            let ext_priv_key = seed.derive_extended_priv_key(bitcoin_network)?;
            let identities = seed.derive_identities();
            let web_password = opts.password.unwrap_or_else(|| seed.derive_auth_password());
            (ext_priv_key, identities, web_password)
        }
    };

    let mut tasks = Tasks::default();

    let (wallet, wallet_feed_receiver) = wallet::Actor::new(network.electrum(), ext_priv_key)?;

    let wallet = wallet.create(None).spawn(&mut tasks);

    if let Some(Withdraw::Withdraw {
        amount,
        address,
        fee,
    }) = network.withdraw()
    {
        wallet
            .send(wallet::Withdraw {
                amount: *amount,
                address: address.clone(),
                fee: fee.map(FeeRate::from_sat_per_vb),
            })
            .await??;

        return Ok(());
    }

    let auth_username = rocket_basicauth::Username("itchysats");
    tracing::info!("Authentication details: username='{auth_username}' password='{web_password}'");

    let figment = rocket::Config::figment()
        .merge(("address", opts.http_address.ip()))
        .merge(("port", opts.http_address.port()))
        .merge(("cli_colors", false));

    let db = db::connect(data_dir.join("taker.sqlite")).await?;

    // Create actors

    let (projection_actor, projection_context) = xtra::Context::new(None);

    let taker = TakerActorSystem::new(
        db.clone(),
        wallet.clone(),
        *olivia::PUBLIC_KEY,
        identities,
        |executor| oracle::Actor::new(db.clone(), executor, SETTLEMENT_INTERVAL),
        {
            |executor| {
                let electrum = network.electrum().to_string();
                monitor::Actor::new(db.clone(), electrum, executor)
            }
        },
        xtra_bitmex_price_feed::Actor::default,
        N_PAYOUTS,
        HEARTBEAT_INTERVAL,
        Duration::from_secs(10),
        projection_actor.clone(),
        maker_identity,
    )?;

    let (proj_actor, projection_feeds) =
        projection::Actor::new(db.clone(), bitcoin_network, &taker.price_feed_actor);
    tasks.add(projection_context.run(proj_actor));

    let possible_addresses = resolve_maker_addresses(maker_url.as_str()).await?;

    tasks.add(connect(
        taker.maker_online_status_feed_receiver.clone(),
        taker.connection_actor.clone(),
        maker_identity,
        possible_addresses,
    ));

    rocket::custom(figment)
        .manage(projection_feeds)
        .manage(wallet_feed_receiver)
        .manage(bitcoin_network)
        .manage(taker.maker_online_status_feed_receiver.clone())
        .manage(taker)
        .manage(auth_username)
        .manage(web_password)
        .mount(
            "/api",
            rocket::routes![
                routes::feed,
                routes::post_order_request,
                routes::get_health_check,
                routes::post_cfd_action,
                routes::post_withdraw_request,
            ],
        )
        .register("/api", rocket::catchers![rocket_basicauth::unauthorized])
        .mount("/", rocket::routes![routes::dist, routes::index])
        .register("/", rocket::catchers![rocket_basicauth::unauthorized])
        .attach(fairings::log_launch())
        .attach(fairings::log_requests())
        .launch()
        .await?;

    db.close().await;

    Ok(())
}

async fn resolve_maker_addresses(maker_addr: &str) -> Result<Vec<SocketAddr>> {
    let possible_addresses = tokio::net::lookup_host(maker_addr)
        .await?
        .collect::<Vec<_>>();

    tracing::debug!(
        "Resolved {} to [{}]",
        maker_addr,
        itertools::join(possible_addresses.iter(), ",")
    );
    Ok(possible_addresses)
}
